import copy
import io
import multiprocessing
import os
import pathlib
import re
import sys
import typing
from argparse import Namespace
from collections import defaultdict
from contextlib import redirect_stdout
from datetime import timedelta
from pprint import pprint
from unittest import mock
from unittest.mock import MagicMock, Mock, patch

# Assuming that the modified _progress_updater function is in a module named runem
import pytest

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import InformativeDict
from runem.runem import (
    MainReturnType,
    _main,
    _process_jobs,
    _process_jobs_by_phase,
    _update_progress,
    timed_main,
)
from runem.types.errors import SystemExitBad
from runem.types.filters import FilePathListLookup
from runem.types.hooks import HookName
from runem.types.runem_config import (
    Config,
    GlobalSerialisedConfig,
    HookSerialisedConfig,
    JobConfig,
    Jobs,
    JobSerialisedConfig,
    PhaseGroupedJobs,
    UserConfigMetadata,
)
from runem.types.types_jobs import JobReturn, JobTiming
from tests.intentional_test_error import IntentionalTestError
from tests.sanitise_reports_footer import sanitise_reports_footer

DIR_REPLACEMENT_DIR: str = "[TEST_REPLACED_DIR]"


def replace_max_cores(stdout: str) -> str:
    """Replace instances of '(of X max)' with '(of [NUM CORES] max)' in a given string.

    Params:
        stdout (str): The stdout to have the non-deterministic 'cores' output replaced.

    Returns:
        str: Modified string with '(of X max)' replaced by '(of [NUM CORES] max)'.
    """
    return re.sub(r"\(of \d+ max\)", "(of [NUM CORES] max)", stdout)


def _strip_reports_footer(runem_stdout: typing.List[str]) -> typing.List[str]:
    idx = runem_stdout.index("runem: reports:")
    return runem_stdout[:idx]


def test_runem_basic() -> None:
    """Tests new user's first call-path, when they wouldn't have a .runem.yml."""
    with io.StringIO() as buf, redirect_stdout(buf):
        with pytest.raises(SystemExit):
            timed_main([])
        runem_stdout = buf.getvalue()

        # this is what we should see when first installing runem
        # TODO: add an on-boarding work flow
        assert "ERROR: Config not found! Looked from" in runem_stdout


@patch(
    "runem.runem.load_user_configs",
    return_value=[],
)
@patch(
    "runem.runem.load_project_config",
)
@patch(
    "runem.runem.find_files",
)
def test_runem_basic_with_config(
    find_files_mock: Mock, load_config_mock: Mock, load_config_metadata_mock: Mock
) -> None:
    global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": ("mock phase",),
            "files": [],
            "options": [],
            "min_version": None,
        }
    }
    empty_config: Config = [
        global_config,
    ]
    minimal_file_lists = defaultdict(list)
    minimal_file_lists["mock phase"].append(pathlib.Path("/test") / "dummy" / "path")
    load_config_mock.return_value = (empty_config, pathlib.Path())
    find_files_mock.return_value = minimal_file_lists
    with io.StringIO() as buf, redirect_stdout(buf):
        # with pytest.raises(SystemExit):
        timed_main(["--help"])
        runem_stdout: typing.List[str] = sanitise_reports_footer(buf.getvalue())
        assert [] == _strip_reports_footer(runem_stdout)

    pprint(runem_stdout)
    assert runem_stdout == [
        "runem: reports:",
        "runem (total wall-clock)  [<float>]",
        "├runem.pre-build          [<float>]",
        "└runem.run-phases         [<float>]",
        " └mock phase (user-time)  [<float>]",
        "runem: DONE: runem took: <float>, saving you <float>, without runem you "
        "would have waited <float>",
        "",
    ]
    load_config_metadata_mock.assert_called_once()


@patch(
    "runem.runem.load_user_configs",
    return_value=[],
)
@patch(
    "runem.runem.load_project_config",
)
@patch(
    "runem.runem.find_files",
)
def test_runem_basic_with_config_no_options(
    find_files_mock: Mock,
    load_config_mock: Mock,
    load_config_metadata_mock: Mock,
) -> None:
    global_config: GlobalSerialisedConfig = {
        "config": {  # type: ignore[typeddict-item]
            "phases": ("mock phase",),
            "files": [],
            # "options": [],
        }
    }
    empty_config: Config = [
        global_config,
    ]
    minimal_file_lists = defaultdict(list)
    minimal_file_lists["mock phase"].append(pathlib.Path("/test") / "dummy" / "path")
    load_config_mock.return_value = (empty_config, pathlib.Path())
    find_files_mock.return_value = minimal_file_lists
    with io.StringIO() as buf, redirect_stdout(buf):
        # with pytest.raises(SystemExit):
        timed_main(["--help"])
        runem_stdout = buf.getvalue()
        assert sanitise_reports_footer(runem_stdout) == [
            "runem: reports:",
            "runem (total wall-clock)  [<float>]",
            "├runem.pre-build          [<float>]",
            "└runem.run-phases         [<float>]",
            " └mock phase (user-time)  [<float>]",
            "runem: DONE: runem took: <float>, saving you <float>, without runem you "
            "would have waited <float>",
            "",
        ]


MOCK_JOB_EXECUTE_INNER_RET: typing.Tuple[JobTiming, JobReturn] = (
    {"job": ("mocked job run", timedelta(0)), "commands": []},
    None,
)

MOCK_USER_LOCAL_HOOK: HookSerialisedConfig = {
    "hook": {
        "command": "echo 'mock user local hook'",
        "hook_name": HookName.ON_EXIT,
    }
}

MOCK_USER_HOME_DIR_HOOK: HookSerialisedConfig = {
    "hook": {
        "command": "echo 'mock user home-dir hook'",
        "hook_name": HookName.ON_EXIT,
    }
}

MOCK_USER_LOCAL_CONFIG: Config = [
    MOCK_USER_LOCAL_HOOK,
]
MOCK_USER_HOME_DIR_CONFIG: Config = [
    MOCK_USER_HOME_DIR_HOOK,
]

MOCK_USER_CONFIGS_METADATA: UserConfigMetadata = [
    (MOCK_USER_LOCAL_CONFIG, pathlib.Path("local-user-config.yml")),
    (MOCK_USER_HOME_DIR_CONFIG, pathlib.Path("local-user-home-dir-config.yml")),
]


@patch(
    "runem.runem.load_user_configs",
    return_value=MOCK_USER_CONFIGS_METADATA,
)
@patch(
    "runem.runem.load_project_config",
)
@patch(
    "runem.runem.find_files",
)
@patch(
    # patch the inner call that is NOT serialised by multiprocessing
    "runem.job_execute.job_execute_inner",
    return_value=MOCK_JOB_EXECUTE_INNER_RET,
)
def _run_full_config_runem(
    job_runner_mock: Mock,
    find_files_mock: Mock,
    load_config_mock: Mock,
    load_config_metadata_mock: Mock,
    runem_cli_switches: typing.List[str],
    add_verbose_switch: bool = True,
    add_silent_switch: bool = False,
    add_command_one_liner: bool = True,
) -> typing.Tuple[typing.List[str], typing.Optional[BaseException]]:
    """A wrapper around running runem e2e tests.

    'runem_cli_switches' should be the runem args, and NOT include the executable at
    index 0.

    Returns a list of lines of terminal output
    """
    global_config: GlobalSerialisedConfig = {
        "config": {
            "phases": ("dummy phase 1", "dummy phase 2"),
            "files": [],
            "min_version": None,
            "options": [
                {
                    "option": {
                        "default": True,
                        "desc": "a dummy option description",
                        "aliases": [
                            "dummy option 1 multi alias 1",
                            "dummy option 1 multi alias 2",
                            "x",
                        ],
                        "alias": "dummy option alias 1",
                        "name": "dummy option 1 - complete option",
                        "type": "bool",
                    }
                },
                {
                    "option": {
                        "default": True,
                        "name": "dummy option 2 - minimal",
                        "type": "bool",
                    }
                },
            ],
        }
    }
    job_config_1: JobSerialisedConfig = {
        "job": {
            "addr": {
                "file": __file__,
                "function": "test_runem_with_full_config",
            },
            "label": "dummy job label 1",
            "when": {
                "phase": "dummy phase 1",
                "tags": set(
                    (
                        "dummy tag 1",
                        "dummy tag 2",
                        "tag only on job 1",
                    )
                ),
            },
        }
    }
    job_config_2: JobSerialisedConfig = {
        "job": {
            "addr": {
                "file": __file__,
                "function": "test_runem_with_full_config",
            },
            "label": "dummy job label 2",
            "when": {
                "phase": "dummy phase 2",
                "tags": set(
                    (
                        "dummy tag 1",
                        "dummy tag 2",
                        "tag only on job 2",
                    )
                ),
            },
        }
    }
    minimal_command_config: JobSerialisedConfig = {
        "job": {
            "command": 'echo "hello world!"',
        }
    }
    command_config: JobSerialisedConfig = {
        "job": {
            "command": 'echo "hello world!"',
            "label": "hello world",
            "when": {
                "phase": "dummy phase 2",
                "tags": set(
                    (
                        "dummy tag 1",
                        "dummy tag 2",
                    )
                ),
            },
        }
    }
    full_config: Config = [
        global_config,
        job_config_1,
        job_config_2,
        command_config,
    ]

    if add_command_one_liner:
        full_config.append(
            minimal_command_config,
        )
    minimal_file_lists = defaultdict(list)
    minimal_file_lists["mock phase"].append(pathlib.Path("/test") / "dummy" / "path")
    mocked_config_path = pathlib.Path(__file__).parent / ".runem.yml"
    load_config_mock.return_value = (full_config, mocked_config_path)
    find_files_mock.return_value = minimal_file_lists
    error_raised: typing.Optional[BaseException] = None
    argv: typing.List[str] = [
        "runem_exec",
        *runem_cli_switches,
    ]
    if add_verbose_switch:
        argv.append("--verbose")
    if add_silent_switch:
        argv.append("--silent")

    with io.StringIO() as buf, redirect_stdout(buf):
        # amend the args to have the exec at 0 as expected by argparse
        try:
            timed_main(argv)
        except BaseException as err:  # pylint: disable=broad-exception-caught
            error_raised = err
        runem_stdout = buf.getvalue()
    # normalise the output
    runem_stdout_lines: typing.List[str] = replace_max_cores(
        # replace the config path as it's different on different systems
        runem_stdout.replace(str(mocked_config_path), "[CONFIG PATH]")
    ).split("\n")
    # job_runner_mock.assert_called()
    got_to_reports: typing.Optional[int] = None
    try:
        got_to_reports = runem_stdout_lines.index("runem: reports:")
    except ValueError:
        pass

    if got_to_reports is not None:
        # truncate the stdout up to where the reports are logged
        runem_stdout_lines = runem_stdout_lines[:got_to_reports]
    return runem_stdout_lines, error_raised


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "silent",
    [
        True,
        False,
    ],
)
def test_runem_with_full_config(verbosity: bool, silent: bool) -> None:
    """End-2-end test with a full config."""
    runem_cli_switches: typing.List[str] = []  # default switches/behaviour
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_verbose_switch=verbosity,
        add_silent_switch=silent,
    )
    if error_raised is not None:
        if not (silent and verbosity):  # pragma: no cover
            # No special handling, this is an unexpected case, so just print the output
            # for debugging, and re-raise.
            print("\n".join(runem_stdout))
            raise error_raised  # re-raise the error that shouldn't have been raised
        assert isinstance(error_raised, SystemExit)
        assert runem_stdout == [
            "runem: cannot parse '--verbose' and '--silent'",
            "",
        ]
        return

    if not verbosity:
        if silent:
            assert runem_stdout == [
                (
                    "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]
        else:
            assert runem_stdout == [
                (
                    "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                    "'dummy phase 1'"
                ),
                (
                    "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]
    else:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                "'tag only on job 1', 'tag only on job 2'"
            ),
            "runem: will run 2 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1', 'echo \"hello world!\"'",
            "runem: will run 2 jobs for phase 'dummy phase 2'",
            "runem:  'dummy job label 2', 'hello world'",
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
            "runem: Running Phase dummy phase 2",
            (
                "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
        ]


def test_runem_with_full_config_verbose() -> None:
    """End-2-end test with a full config."""
    runem_cli_switches: typing.List[str] = ["--verbose"]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        "runem: loaded config from [CONFIG PATH]",
        "runem: found 1 batches, 1 'mock phase' files, ",
        (
            "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
            "'tag only on job 1', 'tag only on job 2'"
        ),
        "runem: will run 2 jobs for phase 'dummy phase 1'",
        "runem:  'dummy job label 1', 'echo \"hello world!\"'",
        "runem: will run 2 jobs for phase 'dummy phase 2'",
        "runem:  'dummy job label 2', 'hello world'",
        "runem: Running Phase dummy phase 1",
        "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) processing 2 jobs",
        "runem: Running Phase dummy phase 2",
        "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) processing 2 jobs",
    ]


def test_runem_with_single_phase() -> None:
    """End-2-end test with a full config choosing only a single phase."""
    runem_cli_switches: typing.List[str] = ["--phases", "dummy phase 1"]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        "runem: loaded config from [CONFIG PATH]",
        "runem: found 1 batches, 1 'mock phase' files, ",
        (
            "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
            "'tag only on job 1', 'tag only on job 2'"
        ),
        "runem: will run 2 jobs for phase 'dummy phase 1'",
        "runem:  'dummy job label 1', 'echo \"hello world!\"'",
        "runem: skipping phase 'dummy phase 2'",
        "runem: Running Phase dummy phase 1",
        (
            "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
            "processing 2 jobs"
        ),
    ]


def test_runem_with_single_phase_verbose() -> None:
    """End-2-end test with a full config choosing only a single phase."""
    runem_cli_switches: typing.List[str] = ["--phases", "dummy phase 1", "--verbose"]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )

    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        "runem: loaded config from [CONFIG PATH]",
        "runem: found 1 batches, 1 'mock phase' files, ",
        (
            "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
            "'tag only on job 1', 'tag only on job 2'"
        ),
        "runem: will run 2 jobs for phase 'dummy phase 1'",
        "runem:  'dummy job label 1', 'echo \"hello world!\"'",
        "runem: skipping phase 'dummy phase 2'",
        "runem: Running Phase dummy phase 1",
        (
            "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
            "processing 2 jobs"
        ),
    ]


def _replace_whitespace_with_new_line(input_string: str) -> str:
    """Replaces all whitespace with a single new line."""
    return re.sub(r"\s+", "\n", input_string)


def _remove_first_line_and_split_along_whitespace(
    input_string: str,
) -> typing.List[str]:
    """Because of how argparse prints help, we need to conform it.

    To conform it we replace all whitespace with a single new-line and then split it
    into a list of strings
    """
    lines: typing.List[str] = input_string.split("\n")

    # remove the usage line as it is something like one of the following,
    # depending on whether we're running in xdist(threaded) or single-threaded
    # contexts:
    # usage: __main__.py [-h] [--jobs JOBS [JOBS ...]]
    # usage: -c [-h] [--jobs JOBS [JOBS ...]]
    index_of_usage_lines: typing.Tuple[int, ...] = (1, 9)
    usage_line: typing.Optional[str] = None
    for index_of_usage_line in index_of_usage_lines:  # pragma: no cover
        usage_line_candidate: str = lines[index_of_usage_line]
        if "usage:" in usage_line_candidate:
            usage_line = usage_line_candidate
            break
    if usage_line is None:  # pragma: no cover
        pprint(lines)
        raise RuntimeError("test setup failure, 'usage' line =NOT found in stdout")
    first_brace: int = usage_line.index("[")
    usage_line = usage_line[first_brace:]
    lines[index_of_usage_line] = usage_line

    first_line_edited: str = "\n".join(lines)
    conformed_whitespace: str = _replace_whitespace_with_new_line(first_line_edited)
    as_list: typing.List[str] = conformed_whitespace.split("\n")
    return as_list


def _conform_help_output(help_output: typing.List[str]) -> str:
    # we have to remove the run-dir for root_dir from the output
    runem_stdout_str: str = (
        "\n".join(help_output)
        .replace(str(pathlib.Path(__file__).parent), DIR_REPLACEMENT_DIR)
        .replace(
            f"({os.cpu_count()} cores available)",
            "([TEST_REPLACED_CORES] cores available)",
        )
        .replace("options:", "[TEST_REPLACED_OPTION_HEADER]")
        .replace("optional arguments:", "[TEST_REPLACED_OPTION_HEADER]")
    )
    assert runem_stdout_str
    return runem_stdout_str


@patch(
    "runem.command_line._get_config_dir",
    return_value=DIR_REPLACEMENT_DIR,
)
def test_runem_help(
    mock_get_config_dir: mock.MagicMock,
) -> None:
    """End-2-end test check that the help-output hasn't *unexpectedly* changed.

    As we build features we want to ensure that the help output stays consistent as we
    leverage the argparse system to generate the help for a specific .runem.yml config
    """
    # Ensure we get fixed-width output.
    os.environ["RUNEM_FIXED_HELP_WIDTH"] = "1"
    runem_cli_switches: typing.List[str] = ["--help"]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_verbose_switch=False,
    )
    assert runem_stdout
    assert error_raised

    runem_stdout_str: str = _conform_help_output(runem_stdout)

    # can't cover code for multiple versions of python
    version_str: str
    if sys.version_info < (3, 11):  # pragma: no cover
        version_str = "3.10"
    else:  # pragma: no cover
        # above 3.11 seems to be stable, for now at least
        version_str = "3.11"
        # version_str = f"{sys.version_info.major}.{sys.version_info.minor}"

    # grab the expected output
    help_dump: pathlib.Path = (
        pathlib.Path(__file__).parent / "data" / f"help_output.{version_str}.txt"
    ).absolute()
    if os.environ.get("RUNEM_TEST_WRITE_HELP", False):  # pragma: no cover
        help_dump.write_text(runem_stdout_str)

    # we have to strip all whitespace as help adapts to the terminal width
    previous_help_content = help_dump.read_text()
    stripped_expected_help_output: typing.List[str] = (
        _remove_first_line_and_split_along_whitespace(previous_help_content)
    )
    stripped_actual_help_output: typing.List[str] = (
        _remove_first_line_and_split_along_whitespace(runem_stdout_str)
    )
    assert stripped_expected_help_output == stripped_actual_help_output, (
        f"{'<' * 100}\n{runem_stdout_str}\n"
        f"{'=' * 100}{previous_help_content}\n"
        f"{'>' * 100}\n"
    )
    mock_get_config_dir.assert_called_once()


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "switch_to_test",
    [
        "--version",
        "-v",
    ],
)
def test_runem_version(switch_to_test: str, verbosity: bool) -> None:
    """End-2-end test check that the --version switch works."""
    runem_cli_switches: typing.List[str] = [
        switch_to_test,
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_command_one_liner=False,
        add_verbose_switch=verbosity,
    )
    assert runem_stdout
    assert error_raised

    # grab the expected output
    version_file: pathlib.Path = (
        pathlib.Path(__file__).parent.parent / "runem" / "VERSION"
    ).absolute()

    expected_version_output: typing.List[str] = [version_file.read_text().strip(), ""]
    if verbosity:
        expected_version_output = [
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'",
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'",
        ] + expected_version_output

    assert runem_stdout == expected_version_output


def test_runem_root_show() -> None:
    """End-2-end test check that the --root-show switch works."""
    switch_to_test: str = "--root-show"
    runem_cli_switches: typing.List[str] = [
        switch_to_test,
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_command_one_liner=False,
        add_verbose_switch=False,
    )
    assert runem_stdout
    assert error_raised

    # In this test the config file is mocked o a dummy file in tests/ .
    runem_checkout_path: pathlib.Path = (pathlib.Path(__file__).parent).absolute()

    expected_version_output: typing.List[str] = [str(runem_checkout_path), ""]

    assert runem_stdout == expected_version_output


@pytest.mark.parametrize(
    "switch_to_test",
    [
        "--jobs",
        "--not-jobs",
    ],
)
def test_runem_bad_validate_switch_jobs(switch_to_test: str) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        switch_to_test,
        "non existent job name",
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    assert error_raised is not None
    assert isinstance(error_raised, SystemExit)
    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        (
            "runem: ERROR: invalid job-name 'non existent job name' for "
            f"{switch_to_test}, choose from one of 'dummy job label 1', "
            "'dummy job label 2', 'echo \"hello world!\"', 'hello world'"
        ),
        "",
    ]


@pytest.mark.parametrize(
    "switch_to_test",
    [
        "--tags",
        "--not-tags",
    ],
)
def test_runem_bad_validate_switch_tags(switch_to_test: str) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        switch_to_test,
        "non existent tag",
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    assert error_raised is not None
    assert isinstance(error_raised, SystemExit)
    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        (
            f"runem: ERROR: invalid tag 'non existent tag' for {switch_to_test}, "
            "choose from one of 'dummy tag 1', 'dummy tag 2', "
            "'tag only on job 1', 'tag only on job 2'"
        ),
        "",
    ]


def test_runem_bad_validate_tag_exclude() -> None:
    """End-2-end test testing that '--not-tags' overrides '--tags'."""
    runem_cli_switches: typing.List[str] = [
        "--tags",
        "dummy tag 1",
        "--not-tags",
        "tag only on job 2",
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_verbose_switch=True,
    )
    assert error_raised is None
    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        "runem: loaded config from [CONFIG PATH]",
        "runem: found 1 batches, 1 'mock phase' files, ",
        (
            "runem: filtering for tags 'dummy tag 1', excluding jobs with tags 'tag "
            "only on job 2'"
        ),
        "runem: will run 2 jobs for phase 'dummy phase 1'",
        "runem:  'dummy job label 1', 'echo \"hello world!\"'",
        (
            "runem: not running job 'dummy job label 2' because it contains the "
            "following tags: 'tag only on job 2'"
        ),
        "runem: will run 1 jobs for phase 'dummy phase 2'",
        "runem:  'hello world'",
        "runem: Running Phase dummy phase 1",
        (
            "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
            "processing 2 jobs"
        ),
        "runem: Running Phase dummy phase 2",
        (
            "runem: Running 'dummy phase 2' with 1 workers (of [NUM CORES] max) "
            "processing 1 jobs"
        ),
    ]


@pytest.mark.parametrize(
    "switch_to_test",
    [
        "--phases",
        "--not-phases",
    ],
)
def test_runem_bad_validate_switch_phases(switch_to_test: str) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        switch_to_test,
        "non existent phase",
    ]
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    assert error_raised is not None
    assert isinstance(error_raised, SystemExit)
    assert runem_stdout == [
        (
            "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
            "'dummy phase 1'"
        ),
        "runem: hooks: loading user hooks from 'local-user-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
        "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
        "runem: hooks: initialising 2 hooks",
        "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
            "user local hook'"
        ),
        (
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
            "user home-dir hook'"
        ),
        f"runem: ERROR: invalid phase 'non existent phase' for {switch_to_test}, "
        "choose from one of 'dummy phase 1', 'dummy phase 2'",
        "",
    ]


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_job_filters_work(verbosity: bool) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        "--jobs",
        "dummy job label 1",
    ]
    if verbosity:
        runem_cli_switches.append("--verbose")
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    if verbosity:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                "'tag only on job 1', 'tag only on job 2'"
            ),
            (
                "runem: not running job 'echo \"hello world!\"' because it isn't "
                "in the list of job names. See --jobs and --not-jobs"
            ),
            "runem: will run 1 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1'",
            (
                "runem: not running job 'dummy job label 2' because it isn't in the list "
                "of job names. See --jobs and --not-jobs"
            ),
            (
                "runem: not running job 'hello world' because it isn't in the list of job "
                "names. See --jobs and --not-jobs"
            ),
            (
                "runem: No jobs for phase 'dummy phase 2' tags 'dummy tag 1', "
                "'dummy tag 2', 'tag only on job 1', "
                "'tag only on job 2'"
            ),
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 1 workers (of [NUM CORES] max) "
                "processing 1 jobs"
            ),
        ]
    else:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                "'tag only on job 1', 'tag only on job 2'"
            ),
            (
                "runem: not running job 'echo \"hello world!\"' because it isn't in "
                "the list of job names. See --jobs and --not-jobs"
            ),
            "runem: will run 1 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1'",
            (
                "runem: not running job 'dummy job label 2' because it isn't in the list of "
                "job names. See --jobs and --not-jobs"
            ),
            (
                "runem: not running job 'hello world' because it isn't in the list of "
                "job names. See --jobs and --not-jobs"
            ),
            (
                "runem: No jobs for phase 'dummy phase 2' tags 'dummy tag 1', "
                "'dummy tag 2', 'tag only on job 1', "
                "'tag only on job 2'"
            ),
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 1 workers (of [NUM CORES] max) "
                "processing 1 jobs"
            ),
        ]


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_tag_filters_work(verbosity: bool) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        "--tags",
        "tag only on job 1",
    ]
    if verbosity:
        runem_cli_switches.append("--verbose")
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    if verbosity:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            "runem: filtering for tags 'tag only on job 1'",
            "runem: will run 2 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1', 'echo \"hello world!\"'",
            (
                "runem: not running job 'dummy job label 2' because it doesn't have "
                "any of the following tags: 'tag only on job 1'"
            ),
            (
                "runem: not running job 'hello world' because it doesn't have any "
                "of the following tags: 'tag only on job 1'"
            ),
            "runem: No jobs for phase 'dummy phase 2' tags 'tag only on job 1'",
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
        ]
    else:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            "runem: filtering for tags 'tag only on job 1'",
            "runem: will run 2 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1', 'echo \"hello world!\"'",
            (
                "runem: not running job 'dummy job label 2' because it doesn't have "
                "any of the following tags: 'tag only on job 1'"
            ),
            (
                "runem: not running job 'hello world' because it doesn't have any of the "
                "following tags: 'tag only on job 1'"
            ),
            "runem: No jobs for phase 'dummy phase 2' tags 'tag only on job 1'",
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
        ]


@pytest.mark.parametrize(
    # whether to allow a one liners that runs for all tags
    "one_liner",
    [
        True,
        False,
    ],
)
@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_tag_out_filters_work(verbosity: bool, one_liner: bool) -> None:
    """End-2-end test failing validation on non existent job-names."""
    runem_cli_switches: typing.List[str] = [
        "--not-tags",
        "tag only on job 1",
    ]
    if verbosity:
        runem_cli_switches.append("--verbose")
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches, add_command_one_liner=one_liner
    )

    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    if one_liner:
        if verbosity:
            # one-liner + verbosity
            assert runem_stdout == [
                (
                    "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                    "'dummy phase 1'"
                ),
                "runem: hooks: loading user hooks from 'local-user-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: initialising 2 hooks",
                "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                    "user local hook'"
                ),
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                    "user home-dir hook'"
                ),
                "runem: loaded config from [CONFIG PATH]",
                "runem: found 1 batches, 1 'mock phase' files, ",
                (
                    "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                    "'tag only on job 2', excluding jobs with tags 'tag only on job 1'"
                ),
                (
                    "runem: not running job 'dummy job label 1' because it contains "
                    "the following tags: 'tag only on job 1'"
                ),
                "runem: will run 1 jobs for phase 'dummy phase 1'",
                "runem:  'echo \"hello world!\"'",
                "runem: will run 2 jobs for phase 'dummy phase 2'",
                "runem:  'dummy job label 2', 'hello world'",
                "runem: Running Phase dummy phase 1",
                (
                    "runem: Running 'dummy phase 1' with 1 workers (of [NUM CORES] max) "
                    "processing 1 jobs"
                ),
                "runem: Running Phase dummy phase 2",
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]
        else:
            # one-liner + no-verbosity
            assert runem_stdout == [
                (
                    "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                    "'dummy phase 1'"
                ),
                "runem: hooks: loading user hooks from 'local-user-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: initialising 2 hooks",
                "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                    "user local hook'"
                ),
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                    "user home-dir hook'"
                ),
                "runem: loaded config from [CONFIG PATH]",
                "runem: found 1 batches, 1 'mock phase' files, ",
                (
                    "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                    "'tag only on job 2', excluding jobs with tags 'tag only on job 1'"
                ),
                (
                    "runem: not running job 'dummy job label 1' because it contains "
                    "the following tags: 'tag only on job 1'"
                ),
                "runem: will run 1 jobs for phase 'dummy phase 1'",
                "runem:  'echo \"hello world!\"'",
                "runem: will run 2 jobs for phase 'dummy phase 2'",
                "runem:  'dummy job label 2', 'hello world'",
                "runem: Running Phase dummy phase 1",
                (
                    "runem: Running 'dummy phase 1' with 1 workers (of [NUM CORES] max) "
                    "processing 1 jobs"
                ),
                "runem: Running Phase dummy phase 2",
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]
    else:
        if verbosity:
            # no-one-liner + verbosity
            assert runem_stdout == [
                "runem: hooks: loading user hooks from 'local-user-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: initialising 2 hooks",
                "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                    "user local hook'"
                ),
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                    "user home-dir hook'"
                ),
                "runem: loaded config from [CONFIG PATH]",
                "runem: found 1 batches, 1 'mock phase' files, ",
                (
                    "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                    "'tag only on job 2', excluding jobs with tags 'tag only on job 1'"
                ),
                (
                    "runem: not running job 'dummy job label 1' because it contains "
                    "the following tags: 'tag only on job 1'"
                ),
                (
                    "runem: No jobs for phase 'dummy phase 1' tags 'dummy tag 1', "
                    "'dummy tag 2', 'tag only on job 2'"
                ),
                "runem: will run 2 jobs for phase 'dummy phase 2'",
                "runem:  'dummy job label 2', 'hello world'",
                "runem: Running Phase dummy phase 2",
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]
        else:
            # no-one-liner + no-verbosity
            assert runem_stdout == [
                "runem: hooks: loading user hooks from 'local-user-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
                "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
                "runem: hooks: initialising 2 hooks",
                "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                    "user local hook'"
                ),
                (
                    "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                    "user home-dir hook'"
                ),
                "runem: loaded config from [CONFIG PATH]",
                "runem: found 1 batches, 1 'mock phase' files, ",
                (
                    "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                    "'tag only on job 2', excluding jobs with tags 'tag only on job 1'"
                ),
                (
                    "runem: not running job 'dummy job label 1' because it contains "
                    "the following tags: 'tag only on job 1'"
                ),
                (
                    "runem: No jobs for phase 'dummy phase 1' tags 'dummy tag 1', "
                    "'dummy tag 2', 'tag only on job 2'"
                ),
                "runem: will run 2 jobs for phase 'dummy phase 2'",
                "runem:  'dummy job label 2', 'hello world'",
                "runem: Running Phase dummy phase 2",
                (
                    "runem: Running 'dummy phase 2' with 2 workers (of [NUM CORES] max) "
                    "processing 2 jobs"
                ),
            ]


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_tag_out_filters_work_all_tags(verbosity: bool) -> None:
    """End-2-end test removing all tags in config.

    # TODO: assert that the tags in the runem_cli_switches match all tags in the config.
    """
    runem_cli_switches: typing.List[str] = [
        "--not-tags",
        "tag only on job 1",
        "tag only on job 2",
        "dummy tag 1",
        "dummy tag 2",
    ]
    if verbosity:
        runem_cli_switches.append("--verbose")
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches,
        add_verbose_switch=False,
        add_command_one_liner=False,
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    if verbosity:
        assert runem_stdout == [
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: excluding jobs with tags 'dummy tag 1', 'dummy tag 2', "
                "'tag only on job 1', 'tag only on job 2'"
            ),
            (
                "runem: not running job 'dummy job label 1' because it doesn't have any of "
                "the following tags: "
            ),
            "runem: No jobs for phase 'dummy phase 1' tags ",
            (
                "runem: not running job 'dummy job label 2' because it doesn't have any of "
                "the following tags: "
            ),
            (
                "runem: not running job 'hello world' because it doesn't have any of the "
                "following tags: "
            ),
            "runem: No jobs for phase 'dummy phase 2' tags ",
        ]
    else:
        assert (
            runem_stdout
            == [
                # "runem: found 1 batches, 1 'mock phase' files, ",
                # (
                #     "runem: excluding jobs with tags 'dummy tag 1', 'dummy tag 2', "
                #     "'tag only on job 1', 'tag only on job 2'"
                # ),
                # "runem: No jobs for phase 'dummy phase 1' tags ",
                # "runem: No jobs for phase 'dummy phase 2' tags ",
            ]
        )


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_runem_phase_filters_work(verbosity: bool) -> None:
    """End-2-end test filtering to a single phase."""
    runem_cli_switches: typing.List[str] = [
        "--phases",
        "dummy phase 1",
    ]
    if verbosity:
        runem_cli_switches.append("--verbose")
    runem_stdout: typing.List[str]
    error_raised: typing.Optional[BaseException]
    (
        runem_stdout,
        error_raised,
    ) = _run_full_config_runem(  # pylint: disable=no-value-for-parameter
        runem_cli_switches=runem_cli_switches
    )
    if error_raised is not None:  # pragma: no cover
        raise error_raised  # re-raise the error that shouldn't have been raised

    if verbosity:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', 'tag only on job 1', "
                "'tag only on job 2'"
            ),
            "runem: will run 2 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1', 'echo \"hello world!\"'",
            "runem: skipping phase 'dummy phase 2'",
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
        ]
    else:
        assert runem_stdout == [
            (
                "runem: WARNING: no phase found for 'echo \"hello world!\"', using "
                "'dummy phase 1'"
            ),
            "runem: hooks: loading user hooks from 'local-user-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: loading user hooks from 'local-user-home-dir-config.yml'",
            "runem: hooks:   added 1 user hooks for 'HookName.ON_EXIT'",
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'mock "
                "user local hook'"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo 'mock "
                "user home-dir hook'"
            ),
            "runem: loaded config from [CONFIG PATH]",
            "runem: found 1 batches, 1 'mock phase' files, ",
            (
                "runem: filtering for tags 'dummy tag 1', 'dummy tag 2', "
                "'tag only on job 1', 'tag only on job 2'"
            ),
            "runem: will run 2 jobs for phase 'dummy phase 1'",
            "runem:  'dummy job label 1', 'echo \"hello world!\"'",
            "runem: skipping phase 'dummy phase 2'",
            "runem: Running Phase dummy phase 1",
            (
                "runem: Running 'dummy phase 1' with 2 workers (of [NUM CORES] max) "
                "processing 2 jobs"
            ),
        ]


class SleepCalledError(ValueError):
    """Thrown when the sleep function is called to stop the infinite loop."""


@pytest.fixture(name="mock_sleep")
def create_mock_print_sleep() -> typing.Generator[typing.Tuple[Mock, Mock], None, None]:
    call_count = 0

    def custom_side_effect(*args: typing.Any, **kwargs: typing.Any) -> float:
        nonlocal call_count
        if call_count < 3:
            call_count += 1
            return 0.1  # Return a valid value for the first call
        raise SleepCalledError("Mocked sleep error on the second call")

    with patch("time.sleep", side_effect=custom_side_effect) as mock_sleep:
        yield mock_sleep


@pytest.mark.parametrize(
    # parametrize the spinner logic so we hit the various states where we use
    # 'halo' and where we don't.
    "show_spinner",
    [
        True,
        False,
    ],
)
def test_progress_updater_with_running_jobs(
    mock_sleep: Mock,
    show_spinner: bool,
) -> None:
    running_jobs: typing.Dict[str, str] = {"job1": "running", "job2": "pending"}
    completed_jobs: typing.Dict[str, str] = {}
    with pytest.raises(SleepCalledError), multiprocessing.Manager() as manager:
        _update_progress(
            "dummy label",
            running_jobs,
            completed_jobs,
            all_jobs=[],
            is_running=manager.Value("b", True),
            num_workers=1,
            show_spinner=show_spinner,
        )
    mock_sleep.assert_called()


def test_progress_updater_with_running_jobs_and_10_jobs(mock_sleep: Mock) -> None:
    running_jobs: typing.Dict[str, str] = {"job1": "running", "job2": "pending"}
    completed_jobs: typing.Dict[str, str] = {}
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_job_config",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "py",
                    "format",
                )
            ),
        },
    }
    all_jobs: Jobs = []
    for idx in range(10):
        job_config = copy.copy(job_config)
        job_config["label"] = f"{job_config['label']} {idx}"
        all_jobs.append(job_config)
    with pytest.raises(SleepCalledError), multiprocessing.Manager() as manager:
        _update_progress(
            "dummy label",
            running_jobs,
            completed_jobs,
            all_jobs=all_jobs,
            is_running=manager.Value("b", True),
            num_workers=1,
            show_spinner=False,
        )
    mock_sleep.assert_called()


def test_progress_updater_without_running_jobs(mock_sleep: Mock) -> None:
    running_jobs: typing.Dict[str, str] = {}
    completed_jobs: typing.Dict[str, str] = {}
    with pytest.raises(SleepCalledError), multiprocessing.Manager() as manager:
        _update_progress(
            "dummy label",
            running_jobs,
            completed_jobs,
            all_jobs=[],
            is_running=manager.Value("b", True),
            num_workers=1,
            show_spinner=False,
        )
    mock_sleep.assert_called()


def test_progress_updater_with_empty_running_jobs(mock_sleep: Mock) -> None:
    running_jobs: typing.Dict[str, str] = {"job1": ""}
    completed_jobs: typing.Dict[str, str] = {}
    with pytest.raises(SleepCalledError), multiprocessing.Manager() as manager:
        _update_progress(
            "dummy label",
            running_jobs,
            completed_jobs,
            all_jobs=[],
            is_running=manager.Value("b", True),
            num_workers=1,
            show_spinner=False,
        )
    mock_sleep.assert_called()


@pytest.mark.parametrize(
    # parametrize the spinner logic so we hit the various states where we use
    # 'halo' and where we don't.
    "show_spinner",
    [
        True,
        False,
    ],
)
def test_progress_updater_with_false(show_spinner: bool) -> None:
    running_jobs: typing.Dict[str, str] = {"job1": ""}
    with multiprocessing.Manager() as manager:
        _update_progress(
            "dummy label",
            running_jobs,
            {},
            [],
            manager.Value("b", False),
            1,
            show_spinner=show_spinner,
        )


def gen_dummy_config_metadata() -> ConfigMetadata:
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=pathlib.Path(__file__),
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            "dummy tag": {
                "tag": "dummy tag",
                "regex": ".*1.txt",  # should match just one file
            }
        },
        hook_manager=MagicMock(),
        jobs=defaultdict(list),
        all_job_names=set(),
        all_job_phases=set(),
        all_job_tags=set(),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=False, procs=1),
        jobs_to_run=set(),  # JobNames,
        phases_to_run=set(),  # JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({"option_on": True, "option_off": False}),  # Options,
    )
    return config_metadata


DUMMY_MAIN_RETURN: MainReturnType = (
    gen_dummy_config_metadata(),  # ConfigMetadata,
    {},  # job_run_metadatas,
    IntentionalTestError(),
)


@patch(
    "runem.runem._main",
    return_value=DUMMY_MAIN_RETURN,
)
def test_runem_re_raises_after_reporting(
    main_mock: Mock,
) -> None:
    with pytest.raises(SystemExitBad):
        timed_main([])


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
@patch("runem.runem._process_jobs", return_value=IntentionalTestError())
def test_process_jobs_by_phase_early_exits_with_exceptions(
    mock_process_jobs: Mock,
    verbosity: bool,
) -> None:
    job_phase_1_raises: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_config",
        },
        "label": "a job that should be attempted",
        "when": {
            "phase": "dummy phase 1",
            "tags": set(
                (
                    "dummy tag 1",
                    "dummy tag 2",
                )
            ),
        },
    }
    job_phase_2_not_executed: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_config",
        },
        "label": "not executed because of previous failure",
        "when": {
            "phase": "dummy phase 1",
            "tags": set(
                (
                    "dummy tag 1",
                    "dummy tag 2",
                )
            ),
        },
    }

    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    jobs_by_phase.update(
        {
            "dummy phase 1": [job_phase_1_raises],
            "dummy phase 2": [job_phase_2_not_executed],
        }
    )

    all_job_names = (
        "not executed because of previous failure",
        "a job that should be attempted",
    )
    all_phase_names = (
        "dummy phase 1",
        "dummy phase 2",
    )
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=pathlib.Path(__file__),
        phases=all_phase_names,
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=jobs_by_phase,
        all_job_names=set(all_job_names),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
        # global_config,
        # job_phase_2_not_executed,
        # job_phase_1_raises,
    )

    config_metadata.set_cli_data(
        args=Namespace(verbose=verbosity, procs=1),
        jobs_to_run=set(all_job_names),  # JobNames,
        phases_to_run=set(all_phase_names),  # JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]

    with io.StringIO() as buf, redirect_stdout(buf):
        error: typing.Optional[BaseException] = _process_jobs_by_phase(
            config_metadata=config_metadata,  # ConfigMetadata,
            file_lists=file_lists,  # FilePathListLookup,
            filtered_jobs_by_phase=jobs_by_phase,  # PhaseGroupedJobs,
            in_out_job_run_metadatas={},  # JobRunMetadatasByPhase,
            show_spinner=False,
        )
        runem_stdout = buf.getvalue().split("\n")
    assert mock_process_jobs.call_count == 1
    assert isinstance(error, IntentionalTestError)
    if not verbosity:
        assert runem_stdout == [
            "",
        ]
    else:
        assert runem_stdout == [
            "runem: Running Phase dummy phase 1",
            "runem: ERROR: running phase dummy phase 1: aborting run",
            "",
        ]


@patch("runem.runem.multiprocessing.Pool", side_effect=IntentionalTestError())
def test_process_jobs_early_exits_with_exceptions(
    mock_process_jobs: Mock,
) -> None:
    job_phase_1_raises: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_config",
        },
        "label": "a job that should be attempted",
        "when": {
            "phase": "dummy phase 1",
            "tags": set(
                (
                    "dummy tag 1",
                    "dummy tag 2",
                )
            ),
        },
    }
    job_phase_2_not_executed: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "test_parse_config",
        },
        "label": "not executed because of previous failure",
        "when": {
            "phase": "dummy phase 1",
            "tags": set(
                (
                    "dummy tag 1",
                    "dummy tag 2",
                )
            ),
        },
    }

    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    jobs_by_phase.update(
        {
            "dummy phase 1": [job_phase_1_raises],
            "dummy phase 2": [job_phase_2_not_executed],
        }
    )

    all_job_names = (
        "not executed because of previous failure",
        "a job that should be attempted",
    )
    all_phase_names = (
        "dummy phase 1",
        "dummy phase 2",
    )
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=pathlib.Path(__file__),
        phases=all_phase_names,
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=jobs_by_phase,
        all_job_names=set(all_job_names),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
        # global_config,
        # job_phase_2_not_executed,
        # job_phase_1_raises,
    )

    config_metadata.set_cli_data(
        args=Namespace(verbose=True, procs=1),
        jobs_to_run=set(all_job_names),  # JobNames,
        phases_to_run=set(all_phase_names),  # JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]

    with io.StringIO() as buf, redirect_stdout(buf):
        error: typing.Optional[BaseException] = _process_jobs(
            config_metadata=config_metadata,  # ConfigMetadata,
            file_lists=file_lists,  # FilePathListLookup,
            in_out_job_run_metadatas=defaultdict(list),  # JobRunMetadatasByPhase,
            phase="dummy phase 1",
            jobs=[job_phase_1_raises],  # Jobs
            show_spinner=False,
        )
        runem_stdout = buf.getvalue().split("\n")
    assert mock_process_jobs.call_count == 1
    assert isinstance(error, IntentionalTestError)
    assert runem_stdout == [
        "runem: Running 'dummy phase 1' with 1 workers (of 1 max) processing 1 jobs",
        "",
    ]


@patch("runem.runem._determine_run_parameters")
@patch(
    "runem.runem.find_files", return_value=[]
)  # testing when find_files returns empty
def test_main_handles_no_files(
    mock_find_files: MagicMock, mock__determine_run_parameters: MagicMock
) -> None:
    class MockCfg:
        def __init__(self) -> None:
            self.cfg_filepath = pathlib.Path(__file__)

    mock__determine_run_parameters.return_value = MockCfg()
    with io.StringIO() as buf, redirect_stdout(buf):
        _main([])
        runem_stdout = buf.getvalue().split("\n")

    mock_find_files.assert_called_once_with(mock__determine_run_parameters.return_value)
    mock__determine_run_parameters.assert_called_once_with([])
    assert runem_stdout == [
        "runem: WARNING: no files found",
        "",
    ]



"""tests/test_yaml_validation.py

Unit‑tests for `runem.yaml_validation.validate_yaml`.
"""

from __future__ import annotations

from typing import Any, List

import pytest

import runem.yaml_validation as uut  # unit under test

# ──────────────────────────────────────────────────────────────────────────────
# Helpers
# ──────────────────────────────────────────────────────────────────────────────


class _DummyErr:  # pylint: disable=too-few-public-methods
    """Lite stand‑in for jsonschema.ValidationError."""

    def __init__(self, path: List[str]) -> None:
        self.path = path
        self.message = "boom"


class _FakeValidator:  # pylint: disable=too-few-public-methods
    """Replaces Draft202012Validator inside the UUT."""

    def __init__(self, errors: List[_DummyErr]) -> None:
        self._errors = errors

    def iter_errors(self, _instance: Any) -> List[_DummyErr]:
        return self._errors


# ──────────────────────────────────────────────────────────────────────────────
# Tests
# ──────────────────────────────────────────────────────────────────────────────


def test_validate_yaml_ok(monkeypatch: pytest.MonkeyPatch) -> None:
    """Returns empty list when validator reports no errors."""
    monkeypatch.setattr(
        uut,
        "Draft202012Validator",
        lambda _schema: _FakeValidator([]),
    )

    result = uut.validate_yaml({"k": "v"}, schema={})
    assert result == []


def test_validate_yaml_two_errors_sorted(monkeypatch: pytest.MonkeyPatch) -> None:
    """≥2 errors are returned in path‑sorted order."""
    errs = [_DummyErr(["z"]), _DummyErr(["a"])]  # unsorted on purpose
    monkeypatch.setattr(
        uut,
        "Draft202012Validator",
        lambda _schema: _FakeValidator(errs),
    )

    result = uut.validate_yaml({"x": 1}, schema={})

    assert [e.path for e in result] == [["a"], ["z"]]
    assert len(result) == 2



import pathlib
import sys
from importlib.util import module_from_spec
from importlib.util import spec_from_file_location as module_spec_from_file_location

from runem.types.errors import FunctionNotFound
from runem.types.runem_config import JobWrapper
from runem.types.types_jobs import JobFunction


def _load_python_function_from_module(
    cfg_filepath: pathlib.Path,
    module_name: str,
    module_file_path: pathlib.Path,
    function_to_load: str,
) -> JobFunction:
    """Given a job-description dynamically loads the test-function so we can call it."""
    # first locate the module relative to the config file
    abs_module_file_path: pathlib.Path = (
        cfg_filepath.parent / module_file_path
    ).absolute()

    # load the function
    try:
        module_spec = module_spec_from_file_location(
            function_to_load, abs_module_file_path
        )
        if not module_spec:
            raise FileNotFoundError()
        if not module_spec.loader:  # pragma: FIXME: add code coverage
            raise FunctionNotFound("unable to load module")
    except FileNotFoundError as err:
        raise FunctionNotFound(
            (
                f"unable to load '{function_to_load}' from '{str(module_file_path)} "
                f"relative to '{str(cfg_filepath)}"
            )
        ) from err

    module = module_from_spec(module_spec)
    if not module:
        raise FunctionNotFound("unable to load module")
    sys.modules[module_name] = module
    try:
        module_spec.loader.exec_module(module)
    except FileNotFoundError as err:
        raise FunctionNotFound(
            (
                f"unable to load '{function_to_load}' from '{str(module_file_path)} "
                f"relative to '{str(cfg_filepath)}"
            )
        ) from err
    try:
        function: JobFunction = getattr(module, function_to_load)
    except AttributeError as err:
        raise FunctionNotFound(
            (
                f"Check that function '[blue]{function_to_load}[/blue]' "
                f"exists in '[blue]{str(module_file_path)}[/blue]' as expected in "
                f"your config at '[blue]{str(cfg_filepath)}[/blue]'"
            )
        ) from err
    return function


def _find_job_module(cfg_filepath: pathlib.Path, module_file_path: str) -> pathlib.Path:
    """Attempts to find the true location of the job-function module."""
    module_path: pathlib.Path = pathlib.Path(module_file_path)

    module_path_candidates = [
        module_path,
        module_path.absolute(),
        (cfg_filepath.parent / module_file_path).absolute(),
    ]
    for module_path in module_path_candidates:
        if module_path.exists():
            break
    if not module_path.exists():
        raise FunctionNotFound(
            (
                f"unable to find test-function module looked in {module_path_candidates} "
                f"running from '{pathlib.Path('.').absolute()}'"
            )
        )
    module_path = module_path.absolute()
    return module_path.relative_to(cfg_filepath.parent.absolute())


def get_job_wrapper_py_func(
    job_wrapper: JobWrapper, cfg_filepath: pathlib.Path
) -> JobFunction:
    """For a job, dynamically loads the associated python job-function.

    Side-effects: also re-addressed the job-config.
    """
    function_to_load: str = job_wrapper["addr"]["function"]
    try:
        module_file_path: pathlib.Path = _find_job_module(
            cfg_filepath, job_wrapper["addr"]["file"]
        )
    except FunctionNotFound as err:
        raise FunctionNotFound(
            (
                "runem failed to find "
                f"job.addr.file '{job_wrapper['addr']['file']}' looking for "
                f"job.addr.function '{function_to_load}'"
            )
        ) from err

    anchored_file_path = cfg_filepath.parent / module_file_path
    assert anchored_file_path.exists(), (
        f"{module_file_path} not found at {anchored_file_path}!"
    )

    module_name = module_file_path.stem.replace(" ", "_").replace("-", "_")

    function = _load_python_function_from_module(
        cfg_filepath, module_name, module_file_path, function_to_load
    )

    # re-write the job-config file-path for the module with the one that worked
    job_wrapper["addr"]["file"] = str(module_file_path)
    return function



import pathlib
import typing

from runem.log import error, log
from runem.types.errors import SystemExitBad
from runem.yaml_utils import load_yaml_object
from runem.yaml_validation import ValidationErrors, validate_yaml


def _load_runem_schema() -> typing.Any:
    """Loads and returns the yaml schema for runem.

    Returns:
        Any: the Draft202012Validator conformant schema.
    """
    schema_path: pathlib.Path = pathlib.Path(__file__).with_name("schema.yml")
    if not schema_path.exists():
        error(
            (
                "runem schema file not found, cannot continue! "
                f"Is the install corrupt? {schema_path}"
            )
        )
        raise SystemExitBad(1)
    schema: typing.Any = load_yaml_object(schema_path)
    return schema


def validate_runem_file(
    cfg_filepath: pathlib.Path,
    all_config: typing.Any,
) -> None:
    """Validates the config Loader object against the runem schema.

    Exits if the files does not validate.
    """
    schema: typing.Any = _load_runem_schema()
    errors: ValidationErrors = validate_yaml(all_config, schema)
    if not errors:
        # aok
        return

    error(f"failed to validate runem config [yellow]{cfg_filepath}[/yellow]")
    for err in errors:
        path = ".".join(map(str, err.path)) or "<root>"
        log(f" [yellow]{path}[/yellow]: {err.message}")
    raise SystemExit("Config validation failed.")



import pathlib
import typing

import yaml


def load_yaml_object(yaml_file: pathlib.Path) -> typing.Any:
    """Loads using full_load, a yaml file.

    This is likely to have safety concerns in non-trusted projects.

    Returns:
        YAML Loader object: the full PyYAML loader object.
    """
    # Do a full, untrusted load of the runem config
    # TODO: work out safety concerns of this
    with yaml_file.open("r+", encoding="utf-8") as file_handle:
        full_yaml_object: typing.Any = yaml.full_load(file_handle)
    return full_yaml_object



import re
import typing
from collections import defaultdict
from datetime import timedelta

from runem.log import log
from runem.types.common import OrderedPhases, PhaseName
from runem.types.types_jobs import (
    JobReturn,
    JobRunMetadatasByPhase,
    JobRunReportByPhase,
    JobRunTimesByPhase,
    JobTiming,
    ReportUrlInfo,
    ReportUrls,
    TimingEntries,
)

try:
    import termplotlib
except ImportError:  # pragma: FIXME: add code coverage
    termplotlib = None


def _align_bar_graphs_workaround(original_text: str) -> str:
    """Module termplotlib doesn't align floats, this fixes that.

    This makes it so we can align the point in the floating point string, without it,
    larger numbers push their bars right, instead of at the same place.
    """
    # Find the maximum width between '[' and '.' characters
    max_width = max(
        int(match.end() - match.start() - 2)
        for match in re.finditer(r"\[.*?(\d+)\.", original_text)
    )

    # Replace each line with aligned numbers
    formatted_text = re.sub(
        r"\[.*?(\d+)\.", lambda m: f"[{m.group(1):>{max_width}}.", original_text
    )

    return formatted_text


def replace_bar_graph_characters(text: str, end_str: str, replace_char: str) -> str:
    """Replaces block characters in lines containing `end_str` with give char.

    Args:
        text (str): Text containing lines of bar-graphs (perhaps)
        end_str (str): If contained by a line, the bar-graph shapes are replaced.
        replace_char (str): The character to replace all bocks with

    Returns:
        str: The modified `text` with block characters replaced on specific
             lines.
    """
    # Define the block character and its light shade replacement
    block_chars = (
        "▏▎▍▋▊▉█▌▐▄▀─"  # Extend this string with any additional block characters you use
        "░·"  # also include the chars we might replace with for special bars
    )

    text_lines: typing.List[str] = text.split("\n")

    # Process each line, replacing block characters if `end_str` is present
    modified_lines = [
        (
            line.translate(str.maketrans(block_chars, replace_char * len(block_chars)))
            if end_str in line
            else line
        )
        for line in text_lines
    ]

    return "\n".join(modified_lines)


def _semi_shade_phase_totals(text: str) -> str:
    light_shade_char = "░"
    return replace_bar_graph_characters(text, "(user-time)", light_shade_char)


def _dot_jobs(text: str) -> str:
    dot_char = "·"
    return replace_bar_graph_characters(text, "(+)", dot_char)


def _plot_times(
    wall_clock_for_runem_main: timedelta,
    phase_run_oder: OrderedPhases,
    timing_data: JobRunTimesByPhase,
) -> typing.Tuple[timedelta, timedelta]:
    """Prints a report to terminal on how well we performed.

    Also calculates the wall-clock time-saved for the user.

    Returns the total system time spent and the time-saved.     (system-time-spent,
    wall-clock-time-saved)
    """
    labels: typing.List[str] = []
    times: typing.List[float] = []

    # Track active processing time for jobs, distinct from wall-clock time (the
    # time the user experiences).
    system_time_spent: timedelta = timedelta()  # init to 0

    for idx, phase in enumerate(phase_run_oder):
        not_last_phase: bool = idx < len(phase_run_oder) - 1
        utf8_phase = " ├" if not_last_phase else " └"
        utf8_phase_group = " │" if not_last_phase else "  "
        # log(f"Phase '{phase}' jobs took:")
        phase_start_idx = len(labels)

        phase_job_times: timedelta = _gen_jobs_report(
            phase,
            labels,
            times,
            utf8_phase_group,
            timing_data[phase],
        )
        labels.insert(phase_start_idx, f"{utf8_phase}{phase} (user-time)")
        times.insert(phase_start_idx, phase_job_times.total_seconds())
        system_time_spent += phase_job_times

    runem_app_timing: typing.List[JobTiming] = timing_data["_app"]
    job_metadata: JobTiming
    for idx, job_metadata in enumerate(reversed(runem_app_timing)):
        last_group: bool = idx == 0  # reverse sorted
        utf8_group = "├" if not last_group else "└"
        job_label, job_time_total = job_metadata["job"]
        labels.insert(0, f"{utf8_group}runem.{job_label}")
        times.insert(0, job_time_total.total_seconds())
    labels.insert(0, "runem (total wall-clock)")
    times.insert(0, wall_clock_for_runem_main.total_seconds())
    if termplotlib:
        fig = termplotlib.figure()
        # cspell:disable-next-line
        fig.barh(
            times,
            labels,
            force_ascii=False,
        )
        shaded_bar_graph: str = _semi_shade_phase_totals(fig.get_string())
        dotted_bar_graph: str = _dot_jobs(shaded_bar_graph)

        # ensure the graphs get aligned nicely.
        final_bar_graph: str = _align_bar_graphs_workaround(dotted_bar_graph)
        print(final_bar_graph)
    else:  # pragma: FIXME: add code coverage
        for job_label, time in zip(labels, times):
            log(f"{job_label}: {time}s")

    wall_clock_time_saved: timedelta = system_time_spent - wall_clock_for_runem_main
    return system_time_spent, wall_clock_time_saved


def _gen_jobs_report(
    phase: PhaseName,
    labels: typing.List[str],
    times: typing.List[float],
    utf8_phase_group: str,
    job_timings: typing.List[JobTiming],
) -> timedelta:
    """Gathers the reports for sub-jobs.

    Split out from _plot_times as the code was getting complex
    """
    job_timing: JobTiming

    # Filter out JobTiming instances with non-zero total_seconds
    non_zero_timing_data: typing.List[JobTiming] = [
        job_timing
        for job_timing in job_timings
        if job_timing["job"][1].total_seconds() != 0
    ]

    job_time_sum: timedelta = timedelta()  # init to 0
    for idx, job_timing in enumerate(non_zero_timing_data):
        not_last: bool = idx < len(non_zero_timing_data) - 1
        utf8_job = "├" if not_last else "└"
        utf8_sub_jobs = "│" if not_last else " "
        job_label, job_time_total = job_timing["job"]
        job_bar_label: str = f"{job_label}"
        labels.append(f"{utf8_phase_group}{utf8_job}{job_bar_label}")
        times.append(job_time_total.total_seconds())
        job_time_sum += job_time_total
        sub_command_times: TimingEntries = job_timing["commands"]

        if len(sub_command_times) <= 1:
            # we only have one or fewer sub-commands, just show the job-time
            continue

        # also print the sub-components of the job as we have more than one
        for idx, (sub_job_label, sub_job_time) in enumerate(sub_command_times):
            sub_utf8 = "├"
            if idx == len(sub_command_times) - 1:
                sub_utf8 = "└"
            labels.append(
                f"{utf8_phase_group}{utf8_sub_jobs}{sub_utf8}{sub_job_label} (+)"
            )
            times.append(sub_job_time.total_seconds())
    return job_time_sum


def _print_reports_by_phase(
    phase_run_oder: OrderedPhases, report_data: JobRunReportByPhase
) -> None:
    """Logs out the reports by grouped by phase."""
    for phase in phase_run_oder:
        report_urls: ReportUrls = report_data[phase]
        job_report_url_info: ReportUrlInfo
        for job_report_url_info in report_urls:
            if not job_report_url_info:
                continue
            log(
                f"report: [blue]{str(job_report_url_info[0])}[/blue]: "
                f"{str(job_report_url_info[1])}"
            )


def report_on_run(
    phase_run_oder: OrderedPhases,
    job_run_metadatas: JobRunMetadatasByPhase,
    wall_clock_for_runem_main: timedelta,
) -> typing.Tuple[timedelta, timedelta]:
    """Generate high-level reports AND prints out any reports returned by jobs.

    IMPORTANT: returns the wall-clock time saved to the user.
    """
    log("reports:")

    # First, collate all data, timing and reports
    timing_data: JobRunTimesByPhase = defaultdict(list)
    report_data: JobRunReportByPhase = defaultdict(list)
    phase: PhaseName
    for phase in job_run_metadatas:
        timing: JobTiming
        reports: JobReturn
        for timing, reports in job_run_metadatas[phase]:
            timing_data[phase].append(timing)
            if reports:
                # the job returned some report urls, record them against the
                # job's phase
                report_data[phase].extend(reports["reportUrls"])

    # Now plot the times on the terminal to give a visual report of the timing.
    time_metrics: typing.Tuple[timedelta, timedelta] = _plot_times(
        wall_clock_for_runem_main=wall_clock_for_runem_main,
        phase_run_oder=phase_run_oder,
        timing_data=timing_data,
    )

    # Penultimate-ly print out the available reports grouped by run-phase.
    _print_reports_by_phase(phase_run_oder, report_data)

    # Return the key metrics for runem, the system vs wall-clock time saved to
    # the user
    # TODO: write this to disk
    return time_metrics



import enum


class HookName(enum.Enum):
    """List supported hooks.

    Todo:
    - before all tasks are run, after config is read
    - BEFORE_ALL = "before-all"
    - after all tasks are done, before reporting
    - AFTER_ALL = "after-all"
    """

    # at exit
    ON_EXIT = "on-exit"



import pathlib
import shutil
import typing

from typing_extensions import Unpack

from runem.log import log
from runem.run_command import RunCommandUnhandledError, run_command
from runem.types import FilePathList, JobKwargs, JobName, JobReturnData, Options


def _job_py_code_ruff_reformat(
    **kwargs: typing.Any,
) -> None:
    """Runs python formatting code in serial order as one influences the other."""
    label: JobName = kwargs["label"]
    options: Options = kwargs["options"]
    python_files: FilePathList = kwargs["file_list"]

    # put into 'check' mode if requested on the command line
    extra_args = []
    if options["check-only"]:
        extra_args.append("--check")

    if not options["ruff"]:
        # Do not run `ruff` if opted-out
        return

    # If ruff is enabled we do NOT run black etc. because ruff does that
    # for us, faster and better.
    ruff_format_cmd = [
        "python3",
        "-m",
        "ruff",
        "format",
        *extra_args,
        *python_files,
    ]
    kwargs["label"] = f"{label} ruff"
    run_command(cmd=ruff_format_cmd, **kwargs)


def _job_py_ruff_lint(
    **kwargs: typing.Any,
) -> None:
    """Runs python formatting code in serial order as one influences the other."""
    label: JobName = kwargs["label"]
    options: Options = kwargs["options"]
    python_files: FilePathList = kwargs["file_list"]

    # try to auto-fix issues (one benefit of ruff over flake8 etc.)
    extra_args = []
    if options["fix"]:
        extra_args.append("--fix")

    if not options["ruff"]:
        # Do not run `ruff` if opted-out
        return

    # If ruff is enabled we do NOT run black etc. because ruff does that
    # for us, faster and better.
    ruff_lint_cmd = [
        "python3",
        "-m",
        "ruff",
        "check",
        *extra_args,
        *python_files,
    ]
    kwargs["label"] = f"{label} ruff"
    run_command(cmd=ruff_lint_cmd, **kwargs)


def _job_py_code_reformat_deprecated(
    **kwargs: Unpack[JobKwargs],
) -> None:
    """Runs python formatting code in serial order as one influences the other."""
    label: JobName = kwargs["label"]
    options: Options = kwargs["options"]
    python_files: FilePathList = kwargs["file_list"]

    # put into 'check' mode if requested on the command line
    extra_args = []
    docformatter_extra_args = [
        "--in-place",
    ]
    if options["check-only"]:
        extra_args.append("--check")
        docformatter_extra_args = []  # --inplace is not compatible with --check

    if options["isort"]:
        isort_cmd = [
            "python3",
            "-m",
            "isort",
            "--profile",
            "black",
            "--treat-all-comment-as-code",
            *extra_args,
            *python_files,
        ]
        kwargs["label"] = f"{label} isort"
        run_command(cmd=isort_cmd, **kwargs)

    if options["black"]:
        black_cmd = [
            "python3",
            "-m",
            "black",
            *extra_args,
            *python_files,
        ]
        kwargs["label"] = f"{label} black"
        run_command(cmd=black_cmd, **kwargs)

    if options["docformatter"]:
        docformatter_cmd = [
            "python3",
            "-m",
            "docformatter",
            "--wrap-summaries",
            "88",
            "--wrap-descriptions",
            "88",
            *docformatter_extra_args,
            *extra_args,
            *python_files,
        ]
        allowed_exits: typing.Tuple[int, ...] = (
            0,  # no work/change required
            3,  # no errors, but code was reformatted
        )
        if options["check-only"]:
            # in check it is ONLY ok if no work/change was required
            allowed_exits = (0,)
        kwargs["label"] = f"{label} docformatter"
        run_command(
            cmd=docformatter_cmd,
            ignore_fails=False,
            valid_exit_ids=allowed_exits,
            **kwargs,
        )


def _job_py_pylint(
    **kwargs: Unpack[JobKwargs],
) -> None:
    python_files: FilePathList = kwargs["file_list"]
    root_path: pathlib.Path = kwargs["root_path"]

    pylint_cfg = root_path / ".pylint.rc"
    if not pylint_cfg.exists():
        raise RuntimeError(f"PYLINT Config not found at '{pylint_cfg}'")

    pylint_cmd = [
        "python3",
        "-m",
        "pylint",
        "-j1",
        "--score=n",
        f"--rcfile={pylint_cfg}",
        *python_files,
    ]
    run_command(cmd=pylint_cmd, **kwargs)


def _job_py_flake8(
    **kwargs: Unpack[JobKwargs],
) -> None:
    python_files: FilePathList = kwargs["file_list"]
    root_path: pathlib.Path = kwargs["root_path"]
    flake8_rc = root_path / ".flake8"
    if not flake8_rc.exists():
        raise RuntimeError(f"flake8 config not found at '{flake8_rc}'")

    flake8_cmd = [
        "python3",
        "-m",
        "flake8",
        *python_files,
    ]
    run_command(cmd=flake8_cmd, **kwargs)


def _job_py_mypy(
    **kwargs: Unpack[JobKwargs],
) -> None:
    python_files: FilePathList = kwargs["file_list"]
    mypy_cmd = ["python3", "-m", "mypy", *python_files]
    output = run_command(cmd=mypy_cmd, **kwargs)
    if "mypy.ini" in output or "Not a boolean:" in output:
        raise RunCommandUnhandledError(f"runem: mypy mis-config detected: {output}")


def _delete_old_coverage_reports(root_path: pathlib.Path) -> None:
    """To avoid false-positives on coverage we delete the coverage report files."""
    old_coverage_report_files: typing.List[pathlib.Path] = list(
        root_path.glob(".coverage_report*")
    )
    for old_coverage_report in old_coverage_report_files:
        old_coverage_report.unlink()


def _job_py_pytest(  # noqa: C901 # pylint: disable=too-many-branches,too-many-statements
    **kwargs: Unpack[JobKwargs],
) -> JobReturnData:
    label: JobName = kwargs["label"]
    options: Options = kwargs["options"]
    procs: int = kwargs["procs"]
    root_path: pathlib.Path = kwargs["root_path"]

    reports: JobReturnData = {"reportUrls": []}
    # TODO: use pytest.ini config pytest
    # pytest_cfg = root_path / ".pytest.ini"
    # assert pytest_cfg.exists()

    if not options["unit-test"]:
        # we've disabled unit-testing on the cli
        return reports

    if options["profile"]:
        raise RuntimeError("not implemented - see run_test.sh for how to implement")

    pytest_path = root_path / "tests"
    assert pytest_path.exists()

    coverage_switches: typing.List[str] = []
    coverage_cfg = root_path / ".coveragerc"
    if options["coverage"]:
        _delete_old_coverage_reports(root_path)
        assert coverage_cfg.exists()
        coverage_switches = [
            "--cov=.",
            f"--cov-config={str(coverage_cfg)}",
            "--cov-append",
            "--no-cov-on-fail",  # do not show coverage terminal report when we fail
            "--cov-fail-under=0",  # we do coverage filing later
        ]

    # TODO: do we want to disable logs on pytest runs?
    # "PYTEST_LOG":"--no-print-logs --log-level=CRITICAL" ;

    threading_switches: typing.List[str] = []
    if procs == -1:
        threading_switches = ["-n", "auto"]

    verbose_switches: typing.List[str] = []
    if "verbose" in kwargs and kwargs["verbose"]:
        verbose_switches = ["-vvv"]

    profile_switches: typing.List[str] = []
    cmd_pytest = [
        "python3",
        "-m",
        "pytest",
        "--color=yes",
        *threading_switches,
        # "-c",
        # str(pytest_cfg),
        *coverage_switches,
        "--failed-first",
        "--exitfirst",
        *profile_switches,
        *verbose_switches,
        str(pytest_path),
    ]

    env_overrides: typing.Dict[str, str] = {}

    kwargs["label"] = f"{label} pytest"
    run_command(
        cmd=cmd_pytest,
        env_overrides=env_overrides,
        **kwargs,
    )

    if options["coverage"]:
        reports_dir: pathlib.Path = root_path / "reports"
        reports_dir.mkdir(parents=False, exist_ok=True)
        coverage_output_dir: pathlib.Path = reports_dir / "coverage_python"
        if coverage_output_dir.exists():
            shutil.rmtree(coverage_output_dir)
        coverage_output_dir.mkdir(exist_ok=True)
        if kwargs["verbose"]:
            log("COVERAGE: Collating coverage")
        # first generate the coverage report for our gitlab cicd
        gen_cobertura_coverage_report_cmd = [
            "python3",
            "-m",
            "coverage",
            "xml",
            "-o",
            str(coverage_output_dir / "cobertura.xml"),
            f"--rcfile={str(coverage_cfg)}",
        ]
        kwargs["label"] = f"{label} coverage cobertura"
        run_command(cmd=gen_cobertura_coverage_report_cmd, **kwargs)

        # then a html report
        gen_html_coverage_report_cmd = [
            "python3",
            "-m",
            "coverage",
            "html",
            f"--rcfile={str(coverage_cfg)}",
        ]
        kwargs["label"] = f"{label} coverage html"
        run_command(cmd=gen_html_coverage_report_cmd, **kwargs)

        # then a standard command-line report that causes the tests to fail.
        gen_cli_coverage_report_cmd = [
            "python3",
            "-m",
            "coverage",
            "report",
            "--fail-under=100",
            f"--rcfile={str(coverage_cfg)}",
        ]
        kwargs["label"] = f"{label} coverage cli"
        report_html = coverage_output_dir / "index.html"
        report_cobertura = coverage_output_dir / "cobertura.xml"
        try:
            run_command(cmd=gen_cli_coverage_report_cmd, **kwargs)
        except BaseException:
            print()
            print(report_html)
            print(report_cobertura)
            raise
        assert coverage_output_dir.exists(), coverage_output_dir
        assert report_html.exists(), report_html
        assert report_cobertura.exists(), report_cobertura
        reports["reportUrls"].append(("coverage html", report_html))
        reports["reportUrls"].append(("coverage cobertura", report_cobertura))
        if kwargs["verbose"]:
            log("COVERAGE: cli output done")
    return reports


def _install_python_requirements(
    **kwargs: Unpack[JobKwargs],
) -> None:
    options: Options = kwargs["options"]
    if not (options["install-deps"]):
        # not enabled
        return
    cmd = [
        "python3",
        "-m",
        "pip",
        "install",
        "-e",
        ".[tests]",
    ]
    run_command(cmd=cmd, **kwargs)



import io
import pathlib
import typing
from argparse import Namespace
from collections import defaultdict
from contextlib import redirect_stdout
from unittest.mock import patch

import pytest

from runem.config_metadata import ConfigMetadata
from runem.hook_manager import HookManager
from runem.informative_dict import InformativeDict
from runem.types.hooks import HookName
from runem.types.runem_config import HookConfig, Hooks, PhaseGroupedJobs

mock_hook_function: HookConfig = {
    "command": "echo 'test hook command'",
    "hook_name": HookName("on-exit"),
}


@pytest.fixture(name="hook_man", autouse=True)
def hook_manager_fixture() -> typing.Generator[HookManager, None, None]:
    """Fixture to clean hooks dictionary before each test."""
    dummy_hooks: Hooks = defaultdict(list)
    hook_manager: HookManager = HookManager(dummy_hooks, verbose=False)
    yield hook_manager


@pytest.fixture(name="mock_timer", autouse=True)
def create_mock_print_sleep() -> typing.Generator[None, None, None]:
    with patch("runem.job_execute.timer", return_value=0.0):  # as mock_timer
        yield


def test_hook_lookup_by_name(hook_man: HookManager) -> None:
    assert hook_man.is_valid_hook_name("on-exit")


def test_hook_lookup_by_name_for_bad_name(hook_man: HookManager) -> None:
    assert not hook_man.is_valid_hook_name("non-existent-hook-name")


def test_register_hook(hook_man: HookManager) -> None:
    """Test registering a hook function."""
    hook_man.register_hook(HookName.ON_EXIT, mock_hook_function, verbose=False)
    assert mock_hook_function in hook_man.hooks_store[HookName.ON_EXIT]


@pytest.mark.parametrize(
    "verbosity",
    [
        True,
        False,
    ],
)
def test_deregister_hook(verbosity: bool, hook_man: HookManager) -> None:
    """Test deregistering a hook function."""
    # run the command and capture output
    with io.StringIO() as buf, redirect_stdout(buf):
        hook_man.register_hook(HookName.ON_EXIT, mock_hook_function, verbose=verbosity)
        hook_man.deregister_hook(
            HookName.ON_EXIT, mock_hook_function, verbose=verbosity
        )
        run_command_stdout = buf.getvalue()
    assert mock_hook_function not in hook_man.hooks_store[HookName.ON_EXIT]

    stdout_lines = run_command_stdout.split("\n")
    if not verbosity:
        assert stdout_lines == [""]
    else:
        assert stdout_lines == [
            "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: echo 'test "
            "hook command'",
            "runem: hooks: deregistered hooks for 'HookName.ON_EXIT', have 0",
            "",
        ]


class IntentionalError(Exception):
    pass


def _dummy_hook(**kwargs: typing.Any) -> None:
    print("mock_hook_function_function called aok")


@pytest.mark.parametrize(
    "do_hooks",
    [
        True,
        False,
    ],
)
def test_invoke_hooks(do_hooks: bool, hook_man: HookManager) -> None:
    """Test invoking hook functions."""
    mock_hook_function_py_callable: HookConfig = {
        "addr": {"file": __file__, "function": "_dummy_hook"},
        "hook_name": HookName("on-exit"),
    }

    mock_hook_function_bash: HookConfig = {
        "command": 'echo "mock_hook_function_bash called aok"',
        "hook_name": HookName("on-exit"),
    }
    hooks: Hooks = defaultdict(list)
    if do_hooks:
        hooks[HookName.ON_EXIT].append(mock_hook_function_py_callable)
        hooks[HookName.ON_EXIT].append(mock_hook_function_bash)

    jobs_by_phase: PhaseGroupedJobs = defaultdict(list)
    all_job_names = ()
    all_phase_names = ()
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=pathlib.Path(__file__),
        phases=all_phase_names,
        options_config=tuple(),
        file_filters={},
        hook_manager=hook_man,
        jobs=jobs_by_phase,
        all_job_names=set(all_job_names),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(),
    )

    config_metadata.set_cli_data(
        args=Namespace(verbose=True, procs=1),
        jobs_to_run=set(all_job_names),  # JobNames,
        phases_to_run=set(all_phase_names),  # JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    # run the command and capture output
    with io.StringIO() as buf, redirect_stdout(buf):
        hook_man.initialise_hooks(hooks, verbose=True)
        hook_man.invoke_hooks(HookName.ON_EXIT, config_metadata)
        run_command_stdout = buf.getvalue()

    run_command_stdout = run_command_stdout.replace(__file__, "<THIS_FILE>")
    expected_stdout: typing.List[str] = [
        "runem: hooks: invoking 0 hooks for 'HookName.ON_EXIT'",
        "runem: hooks: done invoking 'HookName.ON_EXIT'",
        "",
    ]
    if do_hooks:
        expected_stdout = [
            "runem: hooks: initialising 2 hooks",
            "runem: hooks:   initialising 2 hooks for 'HookName.ON_EXIT'",
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 1: "
                "<THIS_FILE>._dummy_hook"
            ),
            (
                "runem: hooks: registered hook for 'HookName.ON_EXIT', have 2: echo "
                '"mock_hook_function_bash called aok"'
            ),
            "runem: hooks: invoking 2 hooks for 'HookName.ON_EXIT'",
            "runem: START: 'HookName.ON_EXIT'",
            "runem: job: running: 'HookName.ON_EXIT'",
            "mock_hook_function_function called aok",
            "runem: job: DONE: 'HookName.ON_EXIT': 0:00:00",
            "runem: START: 'HookName.ON_EXIT'",
            "runem: job: running: 'HookName.ON_EXIT'",
            'runem: running: start: HookName.ON_EXIT: echo "mock_hook_function_bash '
            'called aok"',
            '| HookName.ON_EXIT: "mock_hook_function_bash called aok"',
            "| HookName.ON_EXIT: ",
            'runem: running: done: HookName.ON_EXIT: echo "mock_hook_function_bash called '
            'aok"',
            "runem: job: DONE: 'HookName.ON_EXIT': 0:00:00",
            "runem: hooks: done invoking 'HookName.ON_EXIT'",
            "",
        ]
    assert run_command_stdout.split("\n") == expected_stdout


def test_register_non_existent_hook(hook_man: HookManager) -> None:
    """Test error handling for registering a non-existent hook."""
    with pytest.raises(ValueError) as e:
        hook_man.register_hook(
            "non_existent_hook",  # type: ignore
            mock_hook_function,
            verbose=False,
        )
    assert "Hook non_existent_hook does not exist" in str(e.value)


def test_deregister_non_existent_function(hook_man: HookManager) -> None:
    """Test error handling for deregistering a non-existent function."""
    not_registered: HookConfig = {
        "addr": {"file": __file__, "function": "_dummy_hook"},
        "hook_name": HookName("on-exit"),
    }

    with pytest.raises(ValueError) as e:
        hook_man.deregister_hook(HookName.ON_EXIT, not_registered, verbose=False)
    assert "Function not found in hook" in str(e.value)



import io
import pathlib
import typing
from argparse import Namespace
from collections import defaultdict
from contextlib import redirect_stdout
from datetime import timedelta
from unittest.mock import MagicMock, patch

import pytest

from runem.config_metadata import ConfigMetadata
from runem.informative_dict import InformativeDict
from runem.job_execute import job_execute
from runem.run_command import RecordSubJobTimeType
from runem.types.filters import FilePathListLookup
from runem.types.runem_config import JobConfig, PhaseGroupedJobs
from tests.intentional_test_error import IntentionalTestError


def empty_function(**kwargs: typing.Any) -> None:
    """Does nothing, called by runner."""


def intentionally_raising_function(**kwargs: typing.Any) -> None:
    """Raises an error called by runner."""
    raise IntentionalTestError()


def time_recording_function(
    record_sub_job_time: typing.Optional[RecordSubJobTimeType], **kwargs: typing.Any
) -> None:
    """Does nothing, called by runner."""
    assert record_sub_job_time is not None
    record_sub_job_time("test entry", timedelta(seconds=100))
    record_sub_job_time("test entry 2", timedelta(seconds=200))


def _job_execute_and_capture_stdout(
    job_config: JobConfig,
    running_jobs: typing.Dict[str, str],
    config_metadata: ConfigMetadata,
    file_lists: FilePathListLookup,
) -> typing.Tuple[str, typing.Optional[BaseException]]:
    """Runs job execute and asserts on stdout."""
    ret_err: typing.Optional[BaseException] = None
    with io.StringIO() as buf, redirect_stdout(buf):
        try:
            job_execute(job_config, running_jobs, {}, config_metadata, file_lists)
        except BaseException as err:  # pylint: disable=broad-exception-caught
            # capture the error and return it
            ret_err = err
        # also capture the stdout so we can inspect that as well.
        stdout: str = buf.getvalue()
    # float_less_stdout = re.sub(r"\d+\.\d+", "[FLOAT]", stdout)
    return stdout, ret_err


@pytest.fixture(name="mock_timer", autouse=True)
def create_mock_print_sleep() -> typing.Generator[None, None, None]:
    with patch("runem.job_execute.timer", return_value=0.0):  # as mock_timer
        yield


def test_job_execute_basic_call() -> None:
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "empty_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "dummy tag",
                    "tag not in files",
                )
            ),
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=False, procs=1),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]
    stdout, _ = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    assert stdout == ""


def test_job_execute_basic_call_verbose() -> None:
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "empty_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(("dummy tag",)),
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=True, procs=1),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]
    stdout, _ = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    assert stdout == (
        "runem: START: 'reformat py'\n"
        "runem: job: running: 'reformat py'\n"
        "runem: job: DONE: 'reformat py': 0:00:00\n"
    )


@pytest.mark.parametrize(
    "silent",
    [
        True,
        False,
    ],
)
def test_job_execute_empty_files(silent: bool) -> None:
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "empty_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(("dummy tag",)),
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=(not silent), procs=1, silent=silent),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    # file_lists["dummy tag"] = [__file__]
    stdout, _ = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    if silent:
        assert stdout == ""
    else:
        assert stdout == (
            "runem: START: 'reformat py'\n"
            "runem: WARNING: skipping job 'reformat py', no files for job\n"
            # "runem: job: running: 'reformat py'\n"
            # "runem: job: DONE: 'reformat py': 0:00:00\n"
        )


def test_job_execute_with_ctx_cwd() -> None:
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "empty_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(("dummy tag",)),
        },
        "ctx": {
            # set the cwd
            "cwd": ".",
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=True, procs=1),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]
    stdout, _ = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    assert stdout == (
        "runem: START: 'reformat py'\n"
        "runem: job: running: 'reformat py'\n"
        "runem: job: DONE: 'reformat py': 0:00:00\n"
    )


def test_job_execute_with_raising_func() -> None:
    """Tests that the output is sane on a job-function-raise.

    Aka a job-function throw.

    This test should show that it is the Python exception-handling code that
    repeats output, not the run'em code.

    FIXME: this contains a lot of copy-pasted code from above.
    """
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "intentionally_raising_function",
        },
        "label": "intentionally throwing",
        "when": {
            "phase": "edit",
            "tags": set(("dummy tag",)),
        },
        "ctx": {
            # set the cwd
            "cwd": ".",
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=True, procs=1),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]

    stdout, err = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    assert isinstance(err, IntentionalTestError)
    assert stdout.split("\n") == [
        "runem: START: 'intentionally throwing'",
        "runem: job: running: 'intentionally throwing'",
        "",
        "runem: ERROR: job: job 'intentionally throwing' failed to complete!",
        "",
    ]


def test_job_execute_time_recording_function() -> None:
    job_config: JobConfig = {
        "addr": {
            "file": __file__,
            "function": "time_recording_function",
        },
        "label": "reformat py",
        "when": {
            "phase": "edit",
            "tags": set(
                (
                    "dummy tag",
                    "tag not in files",
                )
            ),
        },
    }
    config_file_path = pathlib.Path(__file__).parent / ".runem.yml"

    expected_jobs: PhaseGroupedJobs = defaultdict(list)
    expected_jobs["dummy phase 1"] = [
        job_config,
    ]
    config_metadata: ConfigMetadata = ConfigMetadata(
        cfg_filepath=config_file_path,
        phases=("dummy phase 1",),
        options_config=tuple(),
        file_filters={
            # "dummy tag": {
            #     "tag": "dummy tag",
            #     "regex": ".*1.txt",  # should match just one file
            # }
        },
        hook_manager=MagicMock(),
        jobs=expected_jobs,
        all_job_names=set(("dummy job label",)),
        all_job_phases=set(("dummy phase 1",)),
        all_job_tags=set(
            (
                "dummy tag 2",
                "dummy tag 1",
            )
        ),
    )
    config_metadata.set_cli_data(
        args=Namespace(verbose=False, procs=1),
        jobs_to_run=set((job_config["label"])),  # JobNames,
        phases_to_run=set(),  # ignored JobPhases,
        tags_to_run=set(),  # ignored JobTags,
        tags_to_avoid=set(),  # ignored  JobTags,
        options=InformativeDict({}),  # Options,
    )

    file_lists: FilePathListLookup = defaultdict(list)
    file_lists["dummy tag"] = [__file__]
    stdout, _ = _job_execute_and_capture_stdout(
        job_config,
        {},
        config_metadata,
        file_lists,
    )
    assert stdout == ""



import typing

from runem.types.common import FilePathList, JobTag


class TagFileFilter(typing.TypedDict):
    tag: JobTag
    regex: str


TagFileFilters = typing.Dict[JobTag, TagFileFilter]
FilePathListLookup = typing.DefaultDict[JobTag, FilePathList]



from runem.informative_dict import InformativeDict, ReadOnlyInformativeDict

OptionName = str
OptionValue = bool
OptionsWritable = InformativeDict[OptionName, OptionValue]
OptionsReadOnly = ReadOnlyInformativeDict[OptionName, OptionValue]
Options = OptionsReadOnly



import pathlib
import sys
import typing

from packaging.version import Version

from runem.config_validate import validate_runem_file
from runem.log import error, log
from runem.runem_version import get_runem_version
from runem.types.runem_config import (
    Config,
    GlobalConfig,
    GlobalSerialisedConfig,
    UserConfigMetadata,
)
from runem.yaml_utils import load_yaml_object

CFG_FILE_YAML = pathlib.Path(".runem.yml")


def _search_up_dirs_for_file(
    start_dir: pathlib.Path, search_filename: typing.Union[str, pathlib.Path]
) -> typing.Optional[pathlib.Path]:
    """Search 'up' from start_dir looking for search_filename."""
    while 1:
        cfg_candidate = start_dir / search_filename
        if cfg_candidate.exists():
            return cfg_candidate
        exhausted_stack: bool = start_dir == start_dir.parent
        if exhausted_stack:
            return None
        start_dir = start_dir.parent


def _search_up_multiple_dirs_for_file(
    start_dirs: typing.Iterable[pathlib.Path],
    search_filename: typing.Union[str, pathlib.Path],
) -> typing.Optional[pathlib.Path]:
    """Same as _search_up_dirs_for_file() but for multiple dir start points."""
    for start_dir in start_dirs:
        found: typing.Optional[pathlib.Path] = _search_up_dirs_for_file(
            start_dir, search_filename
        )
        if found is not None:
            return found
    return None


def _find_config_file(
    config_filename: typing.Union[str, pathlib.Path],
) -> typing.Tuple[typing.Optional[pathlib.Path], typing.Tuple[pathlib.Path, ...]]:
    """Searches up from the cwd for the given config file-name."""
    start_dirs = (pathlib.Path(".").absolute(),)
    cfg_candidate: typing.Optional[pathlib.Path] = _search_up_multiple_dirs_for_file(
        start_dirs, config_filename
    )
    return cfg_candidate, start_dirs


def _find_project_cfg() -> pathlib.Path:
    """Searches up from the cwd for the project .runem.yml config file."""
    cfg_candidate: typing.Optional[pathlib.Path]
    start_dirs: typing.Tuple[pathlib.Path, ...]
    cfg_candidate, start_dirs = _find_config_file(config_filename=CFG_FILE_YAML)

    if cfg_candidate:
        return cfg_candidate

    # error out and exit as we currently require the cfg file as it lists jobs.
    error(f"Config not found! Looked from {start_dirs}")
    sys.exit(1)


def _find_local_configs() -> typing.List[pathlib.Path]:
    """Searches for all user-configs and returns the found ones.

    TODO: add some priorities to the files, such that
            - .runem.local.yml has lowest priority
            - $HOME/.runem.user.yml is applied after .local
            - .runem.user.yml overloads all others
    """
    local_configs: typing.List[pathlib.Path] = []
    for config_filename in (".runem.local.yml", ".runem.user.yml"):
        cfg_candidate: typing.Optional[pathlib.Path]
        cfg_candidate, _ = _find_config_file(config_filename)
        if cfg_candidate:
            local_configs.append(cfg_candidate)

    user_home_config: pathlib.Path = pathlib.Path("~/.runem.user.yml")
    if user_home_config.exists():
        local_configs.append(user_home_config)

    return local_configs


def _conform_global_config_types(
    all_config: Config,
) -> typing.Tuple[Config, typing.Optional[GlobalConfig]]:
    """Ensure that the types match the type-spec."""
    assert isinstance(all_config, list)
    # NOTE: A note of performance. This extra loop over the config should have
    #       minimal impact as the global config should _normally_ be first in
    #       the file.
    global_config: typing.Optional[GlobalConfig] = None
    for idx, config in enumerate(all_config):
        # Notice the 'continue' statement.
        g_config: GlobalSerialisedConfig = config  # type: ignore
        if "config" not in g_config:
            # keep searching
            continue
        global_config = g_config["config"]
        if "phases" in global_config:
            all_config[idx]["config"]["phases"] = tuple(  # type: ignore
                global_config["phases"]
            )
    return all_config, global_config


def load_and_parse_config(cfg_filepath: pathlib.Path) -> Config:
    """For the given config file pass, project or user, load it & parse/conform it."""
    all_config = load_yaml_object(cfg_filepath)
    validate_runem_file(
        cfg_filepath,
        all_config,
    )

    conformed_config: Config
    global_config: typing.Optional[GlobalConfig]
    conformed_config, global_config = _conform_global_config_types(all_config)

    # is the config pinned to a version of runem?
    if (
        global_config is not None
        and ("min_version" in global_config)
        and (global_config["min_version"] is not None)
    ):
        # check that the version of runem is supported by the config file
        runem_version = get_runem_version()
        min_required_version: Version = Version(global_config["min_version"].strip())
        if min_required_version > runem_version:
            log(
                (
                    f".runem.yml config requires runem '{min_required_version}', "
                    f"you have '{runem_version}'. You need to update runem."
                )
            )
            sys.exit(1)
    return conformed_config


def load_project_config() -> typing.Tuple[Config, pathlib.Path]:
    """Finds and loads the .runem.yml file for the current project."""
    cfg_filepath: pathlib.Path = _find_project_cfg()
    conformed_config: Config = load_and_parse_config(cfg_filepath)

    return conformed_config, cfg_filepath


def load_user_configs() -> UserConfigMetadata:
    """Returns the user-local configs, that extend/override runem behaviour."""
    user_configs: typing.List[typing.Tuple[Config, pathlib.Path]] = []
    user_config_paths: typing.List[pathlib.Path] = _find_local_configs()
    for config_path in user_config_paths:
        user_config: Config = load_and_parse_config(config_path)
        user_configs.append((user_config, config_path))
    return user_configs
